@@ -1,9 +1,11 @@
 {-# LANGUAGE PatternGuards #-}
 
 import Control.Applicative
+import Control.Monad
 
 import qualified Data.List as List
 
+import System.Directory
 import System.Environment
 import System.Exit
 import System.FilePath
@@ -21,7 +23,7 @@ srcDir         = "src"
 -- | Checks whether a module is declared (un)safe
 
 unsafeModules :: [FilePath]
-unsafeModules = map toAgdaFilePath
+unsafeModules = map modToFile
   [ "Codata.Musical.Cofin"
   , "Codata.Musical.Colist"
   , "Codata.Musical.Colist.Infinite-merge"
@@ -30,39 +32,45 @@ unsafeModules = map toAgdaFilePath
   , "Codata.Musical.Covec"
   , "Codata.Musical.M"
   , "Codata.Musical.Stream"
-  , "Data.Char.Unsafe"
-  , "Data.Float.Unsafe"
-  , "Data.Nat.Unsafe"
-  , "Data.Nat.DivMod.Unsafe"
-  , "Data.String.Unsafe"
-  , "Data.Word.Unsafe"
   , "Debug.Trace"
   , "Foreign.Haskell"
+  , "Foreign.Haskell.Coerce"
+  , "Foreign.Haskell.Either"
   , "Foreign.Haskell.Maybe"
   , "Foreign.Haskell.Pair"
   , "IO"
   , "IO.Primitive"
-  , "Reflection"
   , "Relation.Binary.PropositionalEquality.TrustMe"
+  , "Text.Pretty.Core"
+  , "Text.Pretty"
   ] where
 
-  toAgdaFilePath :: String -> FilePath
-  toAgdaFilePath name = concat
-    [ "src/"
-    , map (\ c -> if c == '.' then '/' else c) name
-    , ".agda"
-    ]
-
 isUnsafeModule :: FilePath -> Bool
-isUnsafeModule =
-  -- GA 2019-02-24: it is crucial to use an anonymous lambda
-  -- here so that `unsafeModules` is shared between all calls
-  -- to `isUnsafeModule`.
-  \ fp -> unqualifiedModuleName fp == "Unsafe"
-       || fp `elem` unsafeModules
+isUnsafeModule fp =
+    unqualifiedModuleName fp == "Unsafe"
+    || fp `elem` unsafeModules
 
 -- | Checks whether a module is declared as using K
 
+withKModules :: [FilePath]
+withKModules = map modToFile
+  [ "Axiom.Extensionality.Heterogeneous"
+  , "Data.Star.BoundedVec"
+  , "Data.Star.Decoration"
+  , "Data.Star.Environment"
+  , "Data.Star.Fin"
+  , "Data.Star.Pointer"
+  , "Data.Star.Vec"
+  , "Data.String.Unsafe"
+  , "Relation.Binary.HeterogeneousEquality"
+  , "Relation.Binary.HeterogeneousEquality.Core"
+  , "Relation.Binary.HeterogeneousEquality.Quotients.Examples"
+  , "Relation.Binary.HeterogeneousEquality.Quotients"
+  , "Relation.Binary.PropositionalEquality.TrustMe"
+  , "Text.Pretty.Core"
+  , "Text.Pretty"
+  ]
+
 isWithKModule :: FilePath -> Bool
 isWithKModule =
   -- GA 2019-02-24: it is crucial to use an anonymous lambda
@@ -71,31 +79,6 @@ isWithKModule =
   \ fp -> unqualifiedModuleName fp == "WithK"
        || fp `elem` withKModules
 
-  where
-
-  withKModules :: [FilePath]
-  withKModules = map modToFile
-    [ "Axiom.Extensionality.Heterogeneous"
-    , "Data.Char.Unsafe"
-    , "Data.Float.Unsafe"
-    , "Data.Nat.Unsafe"
-    , "Data.Nat.DivMod.Unsafe"
-    , "Data.Star.BoundedVec"
-    , "Data.Star.Decoration"
-    , "Data.Star.Environment"
-    , "Data.Star.Fin"
-    , "Data.Star.Pointer"
-    , "Data.Star.Vec"
-    , "Data.String.Unsafe"
-    , "Data.Word.Unsafe"
-    , "Reflection"
-    , "Relation.Binary.HeterogeneousEquality"
-    , "Relation.Binary.HeterogeneousEquality.Core"
-    , "Relation.Binary.HeterogeneousEquality.Quotients.Examples"
-    , "Relation.Binary.HeterogeneousEquality.Quotients"
-    , "Relation.Binary.PropositionalEquality.TrustMe"
-    ]
-
 unqualifiedModuleName :: FilePath -> String
 unqualifiedModuleName = dropExtension . takeFileName
 
@@ -106,7 +89,6 @@ isLibraryModule f =
   takeExtension f `elem` [".agda", ".lagda"]
   && unqualifiedModuleName f /= "Core"
 
-
 ---------------------------------------------------------------------------
 -- Analysing library files
 
@@ -129,7 +111,7 @@ extractHeader mod = extract
     , (info, d2 : rest) <- span ("-- " `List.isPrefixOf`) ss
     , delimiter d2
     = info
-  extract (d1 : _)
+  extract (d1@(_:_) : _)
     | not (delimiter d1)
     , last d1 == '\r'
     = error $ mod ++ " contains \\r, probably due to git misconfiguration; maybe set autocrf to input?"
@@ -141,11 +123,11 @@ extractHeader mod = extract
 -- | A crude classifier looking for lines containing options & trying to guess
 --   whether the safe file is using either @--guardedness@ or @--sized-types@
 
-data Safety = Unsafe | Safe | SafeGuardedness | SafeSizedTypes
+data Status = Deprecated | Unsafe | Safe | SafeGuardedness | SafeSizedTypes
   deriving (Eq)
 
-classify :: FilePath -> [String] -> Safety
-classify fp ls
+classify :: FilePath -> [String] -> [String] -> Status
+classify fp hd ls
   -- We start with sanity checks
   | isUnsafe && safe          = error $ fp ++ contradiction "unsafe" "safe"
   | not (isUnsafe || safe)    = error $ fp ++ uncategorized "unsafe" "safe"
@@ -153,6 +135,7 @@ classify fp ls
   | isWithK && not withK      = error $ fp ++ missingWithK
   | not (isWithK || withoutK) = error $ fp ++ uncategorized "as relying on K" "without-K"
   -- And then perform the actual classification
+  | deprecated                = Deprecated
   | isUnsafe                  = Unsafe
   | guardedness               = SafeGuardedness
   | sizedtypes                = SafeSizedTypes
@@ -173,6 +156,10 @@ classify fp ls
     withK       = option "--with-K"
     withoutK    = option "--without-K"
 
+    -- based on detected comment in header
+    deprecated  = let detect = List.isSubsequenceOf "This module is DEPRECATED."
+                  in not $ null $ filter detect hd
+
     -- GA 2019-02-24: note that we do not reprocess the whole module for every
     -- option check: the shared @options@ definition ensures we only inspect a
     -- handful of lines (at most one, ideally)
@@ -193,18 +180,26 @@ classify fp ls
 data LibraryFile = LibraryFile
   { filepath   :: FilePath -- ^ FilePath of the source file
   , header     :: [String] -- ^ All lines in the headers are already prefixed with \"-- \".
-  , safety     :: Safety   -- ^ Safety options used by the module
+  , status     :: Status   -- ^ Safety options used by the module
   }
 
 analyse :: FilePath -> IO LibraryFile
 analyse fp = do
   ls <- lines <$> readFileUTF8 fp
+  let hd = extractHeader fp ls
   return $ LibraryFile
     { filepath   = fp
-    , header     = extractHeader fp ls
-    , safety     = classify fp ls
+    , header     = hd
+    , status     = classify fp hd ls
     }
 
+checkFilePaths :: String -> [FilePath] -> IO ()
+checkFilePaths cat fps = forM_ fps $ \ fp -> do
+  b <- doesFileExist fp
+  if b
+    then pure ()
+    else error $ fp ++ " is listed as " ++ cat ++ " but does not exist."
+
 ---------------------------------------------------------------------------
 -- Collecting all non-Core library files, analysing them and generating
 -- 4 files:
@@ -219,12 +214,15 @@ main = do
     [] -> return ()
     _  -> hPutStr stderr usage >> exitFailure
 
+  checkFilePaths "unsafe" unsafeModules
+  checkFilePaths "using K" withKModules
+
   header  <- readFileUTF8 headerFile
   modules <- filter isLibraryModule . List.sort <$>
                find always
                     (extension ==? ".agda" ||? extension ==? ".lagda")
                     srcDir
-  libraryfiles <- mapM analyse modules
+  libraryfiles <- filter ((Deprecated /=) . status) <$> mapM analyse modules
 
   let mkModule str = "module " ++ str ++ " where"
 
@@ -238,7 +236,7 @@ main = do
     unlines [ header
             , "{-# OPTIONS --guardedness --sized-types #-}\n"
             , mkModule safeOutputFile
-            , format $ filter ((Unsafe /=) . safety) libraryfiles
+            , format $ filter ((Unsafe /=) . status) libraryfiles
             ]
 
   let safeGuardednessOutputFile = safeOutputFile ++ "Guardedness"
@@ -246,7 +244,7 @@ main = do
     unlines [ header
             , "{-# OPTIONS --safe --guardedness #-}\n"
             , mkModule safeGuardednessOutputFile
-            , format $ filter ((SafeGuardedness ==) . safety) libraryfiles
+            , format $ filter ((SafeGuardedness ==) . status) libraryfiles
             ]
 
   let safeSizedTypesOutputFile = safeOutputFile ++ "SizedTypes"
@@ -254,7 +252,7 @@ main = do
     unlines [ header
             , "{-# OPTIONS --safe --sized-types #-}\n"
             , mkModule safeSizedTypesOutputFile
-            , format $ filter ((SafeSizedTypes ==) . safety) libraryfiles
+            , format $ filter ((SafeSizedTypes ==) . status) libraryfiles
             ]
 
 -- | Usage info.
